#!/usr/sbin/perl

# Copyright Amit P. Singh, Douglas L. Brutlag, Stanford University.


$PDBPATH = "/db/pdb";
$DSSPPATH = "/db/dssp";

#system commands used by this script:
#you may need to switch to full path names for some shells
$file = "file";
$cat = "cat";
$ln = "ln";
$gunzip = "gunzip";
$uncompress = "uncompress";
$rm = "rm";

#$file = "/usr/bin/file";
#$cat = "/bin/cat";
#$ln = "/bin/ln";
#$gunzip = "/usr/bin/gunzip";
#$uncompress = "/usr/bin/uncompress";
#$rm = "/bin/rm";

open STDOUT, ">/dev/null";
open STDERR, ">/dev/null";

select STDOUT; $| = 1;
select STDERR; $| = 1;

$longpdbname = shift(@ARGV);
$newname = shift(@ARGV);

if (-e $newname) {
  `$rm $newname`;
}

($pdbname) = $longpdbname =~ /^(....)/;

$pdbname =~ tr/A-Z/a-z/;

if ((not defined $pdbname) || (not defined $newname)) {
  exit(0);
}

$myfile = findFile($pdbname);

if ($myfile ne "NOFILE") {
  
  $temp = `$file $myfile`;
  
  @ftype = split /\s+/, $temp;
  
  if ($ftype[1] eq "gzip") {	
    `$cat $myfile | $gunzip > $newname`;
  }
  elsif (($ftype[1] eq "compressed") || ($ftype[1] eq "compress'd")) {	
    `$cat $myfile | $uncompress > $newname`;
  }
  else {
    `$ln -s $myfile $newname`;
  }
}

exit(0);

######################################################################

sub findFile() {
  
  ($two_letter) = $_[0] =~ /^[0-9](..)/;
  $mytemp = getFile("$PDBPATH/$two_letter/$_[0].pdb");
  if ($mytemp eq "NOFILE") {
    $mytemp = getFile("$PDBPATH/$_[0].pdb");
  }
  if ($mytemp eq "NOFILE") {
    $mytemp = getFile("$PDBPATH/pdb$_[0].ent");
  }
  if ($mytemp eq "NOFILE") {
    $mytemp = getFile("$DSSPPATH/$_[0].dssp");
  }
  
  
  return $mytemp;
}


sub getFile() {
  
  $mytemp = $_[0];
  
  if (-e $mytemp) {
    return $mytemp;
  }
  
  $mytemp = "$_[0]".".gz";
  if (-e $mytemp) {
    return $mytemp;
  }
  
  $mytemp = "$_[0]".".Z";
  if (-e $mytemp) {
    return $mytemp;
  }
  
  return "NOFILE";
  
}
    
