/* Copyright Amit P. Singh, Douglas L. Brutlag, Stanford University. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "defines.h"

#ifdef DISPLAYON
#include <GL/gl.h>
#endif

typedef struct stackElementStruct {

  int Rval, Cval;
  struct stackElementStruct* nextPtr;

} stackElement;


typedef struct stackTypeStruct {
  stackElement* topOfStack;
  int stackSize;
} stackType;


typedef struct resultsStruct {
  float score;           
  int numptrs;
  int Rptr[MAXPTRS], Cptr[MAXPTRS];
} results;


typedef struct seqStruct {
  int ss[MAXLENGTH];
  float vector[MAXLENGTH][3];
  float origin[MAXLENGTH][3];
  float end[MAXLENGTH][3];
  float length[MAXLENGTH];
} seq;

static seq seqA, seqB;

int globalCalledAddPair;

char letA_ss[] = "HS";
char letB_ss[] = "HS";
int numA_ss = 2;
int numB_ss = 2;
float penalty_ss[] = {0, 0, 0, 0};

int bestR, bestC, bestOldR, bestOldC;
float bestFullScore = 0.0;

int secondBestR, secondBestC, secondBestOldR, secondBestOldC;
float secondBestScore = 0.0;

int tempTotalPairs = 0;
float highestPossibleScore = 0.0;
float currentD = 0.0;



extern int colA, rowB;     
extern results *data;
extern stackType myStack;
extern float maxScore;    
extern int maxIndex[2];      

extern OBJ *rowObject, *colObject;  

extern int local;
extern void initStackElement(stackElement*);
extern void initstackType();
extern void readStack(stackElement *);
extern void pushStack(stackElement*);
extern stackElement* popStack();
extern int getStackSize();
extern void fixMax(int, int);

extern void init_pairs();
extern void add_pair(Vector, Vector, float);
extern void applyMtoV(float*, float*, float*);
extern void applyRottoV(float*, float*, float*);

extern OBJ* currentObjList[10];
extern int numCurrentObjs;
extern OBJ* ObjList[10];
extern int numObjs;


void initsequences_ss(OBJ*, OBJ*);
void writePair_ss(int, int);
void printAlignment_ss(stackElement *, int);
void computeData_ss(int, int, seq*, seq*);
void computeData_onlyAngles(int, int, seq*, seq*);
float computePenalty_ss(int, int, int, int);
float computeScore_ss(int, int, seq*, seq*);
void align_ss(int);
void report_ss(int);
void printScores_ss();
GLfloat findNearest_ss(OBJ*, int, int, GLfloat*, int*);
float computeAngleScore(int, int, int, int, seq*, seq*);
float computeOriginScore(int, int, int, int, seq*, seq*);
float computeLengthScore(int, int, seq*, seq*);
float testFullAlign(int, int, int, int, seq*, seq*);

void foundHelix(int, int, int, int, int);
void foundSheet(int, int, int, int, int);



void foundHelix(int start, int end, int num, int maxnum, int rc) {

  float x1, y1, z1;
  float x2, y2, z2;


  if (end - start == 3) {
    if (end < maxnum) {
      end++;
    }

    if (start > 0) {
      start--;
    }
  }

  else if (end - start == 4) {
    if (end < maxnum) {
      end++;
    }
    else if (start > 0) {
      start--;
    }
  }

  if (rc == 0) {
    
    x1 = ((colObject->points[start][0])*0.7422 + colObject->points[start+1][0] + 
	  colObject->points[start+2][0] + (colObject->points[start+3][0])*0.7422) * 0.26722;
    y1 = ((colObject->points[start][1])*0.7422 + colObject->points[start+1][1] + 
	  colObject->points[start+2][1] + (colObject->points[start+3][1])*0.7422) * 0.26722;
    z1 = ((colObject->points[start][2])*0.7422 + colObject->points[start+1][2] + 
	  colObject->points[start+2][2] + (colObject->points[start+3][2])*0.7422) * 0.26722;
    
    
    x2 = ((colObject->points[end-1][0])*0.7422 + colObject->points[end-2][0] + 
	  colObject->points[end-3][0] + (colObject->points[end-4][0])*0.7422) * 0.26722;
    y2 = ((colObject->points[end-1][1])*0.7422 + colObject->points[end-2][1] + 
	  colObject->points[end-3][1] + (colObject->points[end-4][1])*0.7422) * 0.26722;
    z2 = ((colObject->points[end-1][2])*0.7422 + colObject->points[end-2][2] + 
	  colObject->points[end-3][2] + (colObject->points[end-4][2])*0.7422) * 0.26722;
  
  
    seqA.ss[num] = 0;
    seqA.origin[num][0] = x1;
    seqA.origin[num][1] = y1;
    seqA.origin[num][2] = z1;

    seqA.end[num][0] = x2;
    seqA.end[num][1] = y2;
    seqA.end[num][2] = z2;
    
    seqA.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqA.vector[num][0] = (x2 - x1)/seqA.length[num];
    seqA.vector[num][1] = (y2 - y1)/seqA.length[num];
    seqA.vector[num][2] = (z2 - z1)/seqA.length[num];


  }

  else {
    x1 = ((rowObject->points[start][0])*0.7422 + rowObject->points[start+1][0] + 
	  rowObject->points[start+2][0] + (rowObject->points[start+3][0])*0.7422) * 0.26722;
    y1 = ((rowObject->points[start][1])*0.7422 + rowObject->points[start+1][1] + 
	  rowObject->points[start+2][1] + (rowObject->points[start+3][1])*0.7422) * 0.26722;
    z1 = ((rowObject->points[start][2])*0.7422 + rowObject->points[start+1][2] + 
	  rowObject->points[start+2][2] + (rowObject->points[start+3][2])*0.7422) * 0.26722;
    
    
    x2 = ((rowObject->points[end-1][0])*0.7422 + rowObject->points[end-2][0] + 
	  rowObject->points[end-3][0] + (rowObject->points[end-4][0])*0.7422) * 0.26722;
    y2 = ((rowObject->points[end-1][1])*0.7422 + rowObject->points[end-2][1] + 
	  rowObject->points[end-3][1] + (rowObject->points[end-4][1])*0.7422) * 0.26722;
    z2 = ((rowObject->points[end-1][2])*0.7422 + rowObject->points[end-2][2] + 
	  rowObject->points[end-3][2] + (rowObject->points[end-4][2])*0.7422) * 0.26722;
  
  
    seqB.ss[num] = 0;
    seqB.origin[num][0] = x1;
    seqB.origin[num][1] = y1;
    seqB.origin[num][2] = z1;

    seqB.end[num][0] = x2;
    seqB.end[num][1] = y2;
    seqB.end[num][2] = z2;
    
    seqB.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqB.vector[num][0] = (x2 - x1)/seqB.length[num];
    seqB.vector[num][1] = (y2 - y1)/seqB.length[num];
    seqB.vector[num][2] = (z2 - z1)/seqB.length[num];

  }

}


void foundSheet(int start, int end, int num, int maxnum, int rc) {

  float x1, y1, z1;
  float x2, y2, z2;


  if (rc == 0) {
    x1 = (colObject->points[start][0] + colObject->points[start+1][0]) * 0.5;
    y1 = (colObject->points[start][1] + colObject->points[start+1][1]) * 0.5;
    z1 = (colObject->points[start][2] + colObject->points[start+1][2]) * 0.5;
    
    
    x2 = (colObject->points[end-1][0] + colObject->points[end-2][0]) * 0.5;
    y2 = (colObject->points[end-1][1] + colObject->points[end-2][1]) * 0.5;
    z2 = (colObject->points[end-1][2] + colObject->points[end-2][2]) * 0.5;
    
    
    seqA.ss[num] = 1;
    seqA.origin[num][0] = x1;
    seqA.origin[num][1] = y1;
    seqA.origin[num][2] = z1;
    
    seqA.end[num][0] = x2;
    seqA.end[num][1] = y2;
    seqA.end[num][2] = z2;

    seqA.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqA.vector[num][0] = (x2 - x1)/seqA.length[num];
    seqA.vector[num][1] = (y2 - y1)/seqA.length[num];
    seqA.vector[num][2] = (z2 - z1)/seqA.length[num];

  }

  else {
    x1 = (rowObject->points[start][0] + rowObject->points[start+1][0]) * 0.5;
    y1 = (rowObject->points[start][1] + rowObject->points[start+1][1]) * 0.5;
    z1 = (rowObject->points[start][2] + rowObject->points[start+1][2]) * 0.5;
    
    
    x2 = (rowObject->points[end-1][0] + rowObject->points[end-2][0]) * 0.5;
    y2 = (rowObject->points[end-1][1] + rowObject->points[end-2][1]) * 0.5;
    z2 = (rowObject->points[end-1][2] + rowObject->points[end-2][2]) * 0.5;
    
    
    seqB.ss[num] = 1;
    seqB.origin[num][0] = x1;
    seqB.origin[num][1] = y1;
    seqB.origin[num][2] = z1;
    
    seqB.end[num][0] = x2;
    seqB.end[num][1] = y2;
    seqB.end[num][2] = z2;


    seqB.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqB.vector[num][0] = (x2 - x1)/seqB.length[num];
    seqB.vector[num][1] = (y2 - y1)/seqB.length[num];
    seqB.vector[num][2] = (z2 - z1)/seqB.length[num];

  }
}


void initsequences_ss(OBJ *o1, OBJ *o2) {

  char ch = '0', ch1, ch2;
  char *seq1, *seq2;
  int i;

  globalCalledAddPair = 0;

  colObject = o1;
  rowObject = o2;

  seq1 = colObject->sstype;
  seq2 = rowObject->sstype;

  colA = colObject->ssArrayLength;
  rowB = rowObject->ssArrayLength;

  
  for (i = 0; i < colA; i++) {

    ch = seq1[colObject->ssArray[i*2]];
      
    if (ch == 'H') {
      foundHelix(colObject->ssArray[i*2], colObject->ssArray[i*2 + 1], i, colObject->numPts, 0);
    }
    else if (ch == 'S') {
      foundSheet(colObject->ssArray[i*2], colObject->ssArray[i*2 + 1], i, colObject->numPts, 0);
    }
        
  }


  for (i = 0; i < rowB; i++) {

    ch = seq2[rowObject->ssArray[i*2]];
      
    if (ch == 'H') {
      foundHelix(rowObject->ssArray[i*2], rowObject->ssArray[i*2 + 1], i, rowObject->numPts, 1);
    }
    else if (ch == 'S') {
      foundSheet(rowObject->ssArray[i*2], rowObject->ssArray[i*2 + 1], i, rowObject->numPts, 1);
    }
  }


  data = (results*) malloc((sizeof(results))*(colA+1)*(rowB+1));

  for (i = 0; i < (colA+1)*(rowB+1); i++) {
    data[i].score = -1000000.0;
    data[i].numptrs = 0;
  }

  maxScore = -1000000.0;
  bestFullScore = -100000.0;
  secondBestScore = -100000.0;
  tempTotalPairs = 0;

  highestPossibleScore = (colA < rowB) ? (40.0 * (float)(colA)) : (40.0 * (float)(rowB));

  currentD = 10.0;

  bestR = 1;
  bestC = 1;
  bestOldR = 0;
  bestOldC = 0;

  secondBestR = 1;
  secondBestC = 1;
  secondBestOldR = 0;
  secondBestOldC = 0;

  initstackType();


}


void writePair_ss(int r, int c) {

  int i, myindex, alignSize;
  stackElement *newElement;
  
  newElement = (stackElement*) malloc(sizeof(stackElement));
  initStackElement(newElement);
  newElement->Rval = r;
  newElement->Cval = c;

  pushStack(newElement);
  
  myindex = c + r*colA;
  
  for (i = 0; i < data[myindex].numptrs; i++) {
    writePair_ss(data[myindex].Rptr[i], data[myindex].Cptr[i]);
  }

  alignSize = getStackSize();
  
  if ((alignSize > 0) && (data[myindex].numptrs == 0)) {
    newElement = (stackElement*) malloc((sizeof(stackElement))*alignSize);
    for (i = 0; i < alignSize; i++) 
      initStackElement(&(newElement[i]));
    
    readStack(newElement);

    printAlignment_ss(newElement, alignSize);
    
    free(newElement);
  }

  newElement = popStack();

  free(newElement);

}

GLfloat findNearest_ss(OBJ* target, int targetStart, int window, GLfloat* pVec, int *targetNum) {

  int i, num, start, end;
  GLfloat min = 10000.0;
  GLfloat dist;
  
  for (i = targetStart; i < targetStart + window; i++) {
    
    dist = sqrt((target->points[i][0] - pVec[0])*(target->points[i][0] - pVec[0]) +
		(target->points[i][1] - pVec[1])*(target->points[i][1] - pVec[1]) +
		(target->points[i][2] - pVec[2])*(target->points[i][2] - pVec[2]));
    
    if (dist < min) {
      min = dist;
      num = i;
    }
  }
  
  *targetNum = num;
  
  return min;
  
}


void printAlignment_ss(stackElement *list, int listSize) {

  int i, j, r, c, oldr, oldc, outIndex, tempIndex;
  char *out1, *out2;
  int ii, targetstart, targetend, querystart, queryend, difflength, queryPoint, count;
  GLfloat tdist;

#if SHOWOUTPUT 
  out1 = (char*) malloc((sizeof(char))*(colA + rowB + 20));
  out2 = (char*) malloc((sizeof(char))*(colA + rowB + 20));
  outIndex = 0;
 
  r = list[0].Rval;
  oldr = r - 1;
  c = list[0].Cval;
  oldc = c - 1;

  for (i = 0; i < r - c; i++) {
    out1[outIndex] = '-';
    outIndex++;
  }
  for (i = 0; i < c; i++) {
    out1[outIndex] = letA_ss[seqA.ss[i]];
    outIndex++;
  }
  
  outIndex = 0;
  for (i = 0; i < c - r; i++) {
    out2[outIndex] = '-';
    outIndex++;
  }
  for (i = 0; i < r; i++) {
    out2[outIndex] = letB_ss[seqB.ss[i]];
    outIndex++;
  }

  if (local) {
    out1[outIndex] = '*';
    out2[outIndex] = '*';
    outIndex++;
  }
#endif
  for (i = 0; i < listSize; i++) {
    r = list[i].Rval;
    c = list[i].Cval;

#if SHOWOUTPUT 
    for (j = oldr + 1; j < r; j++) {	
      out1[outIndex] = '-';
      out2[outIndex] = letB_ss[seqB.ss[j]];
      outIndex++;
    }
    for (j = oldc + 1; j < c; j++) {	
      out1[outIndex] = letA_ss[seqA.ss[j]];
      out2[outIndex] = '-';
      outIndex++;
    }
    if (seqA.ss[c] == seqB.ss[r]) {
      out1[outIndex] = letA_ss[seqA.ss[c]];
      out2[outIndex] = letB_ss[seqB.ss[r]];
      outIndex++;
    }
    oldr = r;
    oldc = c; 
#endif
    if (seqA.ss[c] == seqB.ss[r]) {

      globalCalledAddPair++;
	
      targetstart = rowObject->ssArray[r*2];
      targetend = rowObject->ssArray[r*2 + 1];

      querystart = colObject->ssArray[c*2];
      queryend = colObject->ssArray[c*2 + 1];

      difflength = (queryend - querystart) - (targetend - targetstart);

      if (difflength >= 0) {
	count = 0;
	for (ii = targetstart; ii < targetend; ii++) {
	  tdist = findNearest_ss(colObject, querystart + count, difflength+1, 
				 rowObject->points[ii], &queryPoint);
	  add_pair(colObject->points[queryPoint], rowObject->points[ii], 1); 
	  globalCalledAddPair++;
	  
	  count++;
	}

	ii = targetend;
	
	if ((targetend - targetstart == 3) || (queryend - querystart == 3)) {
	  if ((queryPoint+1 < colObject->numPts) && (ii < rowObject->numPts)) {
	    globalCalledAddPair++;
	    add_pair(colObject->points[queryPoint+1], rowObject->points[ii], 1); 
	  }
	  else {
	    globalCalledAddPair++;
	    add_pair(colObject->points[queryPoint], rowObject->points[ii-1], 1); 
	  }
	}


      }

      else if (difflength < 0) {
	difflength = -1.0 * difflength;
	count = 0;
	for (ii = querystart; ii < queryend; ii++) {
	  tdist = findNearest_ss(rowObject, targetstart + count, difflength+1, 
				 colObject->points[ii], &queryPoint);
	  add_pair(colObject->points[ii], rowObject->points[queryPoint], 1); 
	  globalCalledAddPair++;
	  
	  count++;
	}

	ii = queryend;
	
	if ((targetend - targetstart == 3) || (queryend - querystart == 3)) {
	  if ((queryPoint+1 < rowObject->numPts) && (ii < colObject->numPts)) {
	    globalCalledAddPair++;
	    add_pair(colObject->points[ii], rowObject->points[queryPoint+1], 1); 
	  }
	  else {
	    globalCalledAddPair++;
	    add_pair(colObject->points[ii-1], rowObject->points[queryPoint], 1); 
	  }
	}


      } 
    }
  }
#if SHOWOUTPUT 
  if (local) {
    out1[outIndex] = '*';
    out2[outIndex] = '*';
    outIndex++;
  }

  tempIndex = outIndex;
  for (i = c + 1; i < colA; i++) {
    out1[outIndex] = letA_ss[seqA.ss[i]];
    outIndex++;
  }
  for (i = 0; i < (rowB - r) - (colA - c); i++) {
    out1[outIndex] = '-';
    outIndex++;
  }

  outIndex = tempIndex;
  for (i = r + 1; i < rowB; i++) {
    out2[outIndex] = letB_ss[seqB.ss[i]];
    outIndex++;
  }
  for (i = 0; i < (colA - c) - (rowB - r); i++) {
    out2[outIndex] = '-';
    outIndex++;
  }

  out1[outIndex] = '\0';
  out2[outIndex] = '\0'; 

  printf("%s\n", out1);
  printf("%s\n", out2);

  printf("\n");


  free(out1);
  free(out2);
#endif
}

float computeAngleScore(int r, int c, int oldr, int oldc, seq *myseqA, seq *myseqB) {
  
  float retvalue, theta, phi;

  if (myseqA->ss[oldc] != myseqB->ss[oldr]) {
    retvalue = 0.0;
  }
  else {
    theta = (float)(acos((double)((myseqA->vector[c][0])*(myseqA->vector[oldc][0]) + 
				  (myseqA->vector[c][1])*(myseqA->vector[oldc][1]) + 
				  (myseqA->vector[c][2])*(myseqA->vector[oldc][2]))));
    
    phi = (float)(acos((double)((myseqB->vector[r][0])*(myseqB->vector[oldr][0]) + 
				(myseqB->vector[r][1])*(myseqB->vector[oldr][1]) + 
				(myseqB->vector[r][2])*(myseqB->vector[oldr][2]))));
    
    
    if (theta > phi) {
      theta = theta - phi;
    }
    else {
      theta = phi - theta;
    }
    
    
    retvalue = (4.0*(theta))/PI;
    retvalue = retvalue*retvalue;
    
    retvalue = (28.0/(1.0 + retvalue)) - 14.0;
  }

  return retvalue;

}


float computeOriginScore(int r, int c, int oldr, int oldc, seq *myseqA, seq *myseqB) {
  
  float retvalue, colDist, rowDist, x1, y1, z1, theta, theta2, phi, phi2;

  if (myseqA->ss[oldc] != myseqB->ss[oldr]) {
    retvalue = 0.0;
  }
  else {

    x1 = myseqA->origin[c][0] - myseqA->end[oldc][0];
    y1 = myseqA->origin[c][1] - myseqA->end[oldc][1];
    z1 = myseqA->origin[c][2] - myseqA->end[oldc][2];

    colDist = sqrt((x1*x1) + (y1*y1) + (z1*z1));

    x1 = x1/colDist;
    y1 = y1/colDist;
    z1 = z1/colDist;

    theta = (float)(acos((double)((myseqA->vector[oldc][0])*(x1) + 
				  (myseqA->vector[oldc][1])*(y1) + 
				  (myseqA->vector[oldc][2])*(z1))));
    
    phi = (float)(acos((double)((myseqA->vector[c][0])*(x1) + 
				(myseqA->vector[c][1])*(y1) + 
				(myseqA->vector[c][2])*(z1))));
    
        
    x1 = myseqB->origin[r][0] - myseqB->end[oldr][0];
    y1 = myseqB->origin[r][1] - myseqB->end[oldr][1];
    z1 = myseqB->origin[r][2] - myseqB->end[oldr][2];

    rowDist = sqrt((x1*x1) + (y1*y1) + (z1*z1));

    x1 = x1/rowDist;
    y1 = y1/rowDist;
    z1 = z1/rowDist;


    theta2 = (float)(acos((double)((myseqB->vector[oldr][0])*(x1) + 
				  (myseqB->vector[oldr][1])*(y1) + 
				  (myseqB->vector[oldr][2])*(z1))));
    
    phi2 = (float)(acos((double)((myseqB->vector[r][0])*(x1) + 
				(myseqB->vector[r][1])*(y1) + 
				(myseqB->vector[r][2])*(z1))));
    

    
    if (colDist > rowDist) {
      colDist = colDist - rowDist;
    }
    else {
      colDist = rowDist - colDist;
    }
    

    retvalue = colDist/5.0;
    retvalue = retvalue*retvalue;
    
    retvalue = (4.0/(1.0 + retvalue)) - 2.0;


    if (theta > theta2) {
      theta = theta - theta2;
    }
    else {
      theta = theta2 - theta;
    }
    
    
    theta = (4.0*(theta))/PI;
    theta = theta*theta;
    
    retvalue += (4.0/(1.0 + theta)) - 2.0;

    if (phi > phi2) {
      phi = phi - phi2;
    }
    else {
      phi = phi2 - phi;
    }
    
    
    phi = (4.0*(phi))/PI;
    phi = phi*phi;
    
    retvalue += (4.0/(1.0 + phi)) - 2.0;    

  }

  return retvalue;

}

float computeLengthScore(int r, int c, seq *myseqA, seq *myseqB) {

  float retvalue, temp;

  if (myseqA->length[c] > myseqB->length[r]) {
    temp = ((myseqA->length[c] - myseqB->length[r])/7.0);
    temp = temp*temp;
    retvalue = (10.0/(1.0 + temp)) - 5.0;
  }
  else {
    temp = ((myseqB->length[r] - myseqA->length[c])/7.0);
    temp = temp*temp;
    retvalue = (10.0/(1.0 + temp)) - 5.0;
  }

  return retvalue;
}


float computeScore_ss(int r, int c, seq *myseqA, seq *myseqB) {

  float retvalue;
  float x1, y1, z1, temp;

  if (myseqA->ss[c] != myseqB->ss[r]) {
    retvalue = -100000.0;
  }
  else {
    
    x1 = (float)((myseqA->vector[c][0])*(myseqB->vector[r][0]) + 
		 (myseqA->vector[c][1])*(myseqB->vector[r][1]) + 
		 (myseqA->vector[c][2])*(myseqB->vector[r][2]));

    if (x1 > 1.0) {
      x1 = 1.0;
    }

    x1 = (float)(acos((double)(x1)));

    temp = (4.0*x1)/PI;
    temp = temp*temp;

    retvalue = (20.0/(1.0 + temp)) - 10.0;

  }
  
  return retvalue;
}


float computeAvgDistScore_ss(int r, int c, seq *myseqA, seq *myseqB) {

  float retvalue, dist;
  float x1, y1, z1, temp;

  if (myseqA->ss[c] != myseqB->ss[r]) {
    retvalue = -100000.0;
  }
  else {
    
    x1 = myseqA->origin[c][0] - myseqB->origin[r][0];
    y1 = myseqA->origin[c][1] - myseqB->origin[r][1];
    z1 = myseqA->origin[c][2] - myseqB->origin[r][2];

    dist = sqrt((x1*x1) + (y1*y1) + (z1*z1));

    x1 = myseqA->end[c][0] - myseqB->end[r][0];
    y1 = myseqA->end[c][1] - myseqB->end[r][1];
    z1 = myseqA->end[c][2] - myseqB->end[r][2];

    dist =+ sqrt((x1*x1) + (y1*y1) + (z1*z1));


    x1 = ((myseqA->origin[c][0] + myseqA->end[c][0])/2.0) - ((myseqB->origin[r][0] + myseqB->end[r][0])/2.0);
    y1 = ((myseqA->origin[c][1] + myseqA->end[c][1])/2.0) - ((myseqB->origin[r][1] + myseqB->end[r][1])/2.0);
    z1 = ((myseqA->origin[c][2] + myseqA->end[c][2])/2.0) - ((myseqB->origin[r][2] + myseqB->end[r][2])/2.0);

    dist =+ sqrt((x1*x1) + (y1*y1) + (z1*z1));


    dist = dist/3.0;

    retvalue = (10.0/(1.0 + (dist*dist)/16.0)) - 5.0;

  }
  
  return retvalue;
}


void computeData_ss(int r, int c, seq *myseqA, seq *myseqB) {
  int myindex, i, tempNum;
  float tempScore, myscore, max;
  float myscore1, myscore2, myconstscore1, myconstscore2, myconstscore3;


  max = -10000.0;
  
  if (local) 
    max = 0.0;

  myindex = c + r*colA;
  myconstscore1 = computeScore_ss(r, c, myseqA, myseqB);
  myconstscore2 = computeAvgDistScore_ss(r, c, myseqA, myseqB);
  myconstscore3 = computeLengthScore(r, c, myseqA, myseqB); 

  if (myconstscore2 > 0) {    
    myscore = myconstscore1 + myconstscore2  + myconstscore3;
  }
  else {
    myscore = -10000;
  }
  

  if ((r > 0) && (c > 0) && ((myscore > 0) || (local == 0))) {

    for (i = 0; i < r; i++) {

      myscore1 = computeAngleScore(r,c,i,c-1, myseqA, myseqB);
      myscore2 = computeOriginScore(r,c,i,c-1, myseqA, myseqB);


      if ((myscore1 + myscore2 > 0) || (local == 0)) {
	tempScore = data[(c-1) + i*colA].score + myscore + myscore1 + myscore2;
      }
      else {
	tempScore = -10000.0;
      }

      if (tempScore > max) {
	max = tempScore;	
	if ((local == 0) || (data[(c-1) + i*colA].score > 0.0)) {
	  data[myindex].Rptr[0] = i;
	  data[myindex].Cptr[0] = c - 1;
	  data[myindex].numptrs = 1; 
	}
	else {
	  data[myindex].numptrs = 0; 
	}	  
      }
    }

    for (i = 0; i < c - 1; i++) {

      myscore1 = computeAngleScore(r,c,r-1,i, myseqA, myseqB);
      myscore2 = computeOriginScore(r,c,r-1,i, myseqA, myseqB);

      if ((myscore1 + myscore2 > 0) || (local == 0)) {
	tempScore = data[i + (r-1)*colA].score + myscore + myscore1 + myscore2;
      }
      else {
	tempScore = -10000.0;
      }

      if (tempScore > max) {
	max = tempScore;	
	if ((local == 0) || (data[i + (r-1)*colA].score > 0.0)) {
	  data[myindex].Rptr[0] = r - 1;
	  data[myindex].Cptr[0] = i;
	  data[myindex].numptrs = 1;
	}
	else {
	  data[myindex].numptrs = 0;
	}
      }
    }
  }
  else if ((r == 0) || (c == 0)) {
    if (myscore > max)
      max = myscore;
  }

  data[myindex].score = max;	

}


float testFullAlign(int newr, int newc, int oldr, int oldc, seq *myseqA, seq *myseqB) {

  seq newseqB;

  Quaternion quat_rot;
  Vector translation;
  Matrix mat;
  float m[16];

  int i,j,r,c;
  
  tempTotalPairs++;

  for (i = 0; i < (colA+1)*(rowB+1); i++) {
    data[i].score = -1000000.0;
    data[i].numptrs = 0;
  }

  maxScore = -1000000.0;

  initstackType();


  init_pairs();

  add_pair(myseqA->origin[newc], myseqB->origin[newr], 1);
  add_pair(myseqA->end[newc], myseqB->end[newr], 1);
  add_pair(myseqA->origin[oldc], myseqB->origin[oldr], 1);
  add_pair(myseqA->end[oldc], myseqB->end[oldr], 1);

  match_pairs (mat, quat_rot, translation);

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      m[i*4 + j] = mat[i][j];
    }
  }


  for (r = 0; r < rowB; r++) {
    newseqB.ss[r] = myseqB->ss[r];
    newseqB.length[r] = myseqB->length[r];
    applyRottoV(newseqB.vector[r], myseqB->vector[r], m);
    applyMtoV(newseqB.origin[r], myseqB->origin[r], m);
    applyMtoV(newseqB.end[r], myseqB->end[r], m);

  }


  for (r = 0; r < rowB; r++) {
    for (c = 0; c < colA; c++) {
      computeData_ss(r, c, myseqA, &newseqB);
      fixMax(r, c);
    }
  }

  return maxScore;

}




void computeData_onlyAngles(int r, int c, seq *myseqA, seq *myseqB) {
  int i;
  float fullScore;
  
  if ((myseqA->ss[c] == myseqB->ss[r]) && (r > 0) && (c > 0)) {
      
    for (i = 0; i < r; i++) {
      
      if (myseqA->ss[c-1] == myseqB->ss[i]) {
	
	fullScore = computeAngleScore(r,c,i,c-1, myseqA, myseqB) + computeOriginScore(r,c,i,c-1, myseqA, myseqB); 
	if (fullScore > currentD) {
	  
	  fullScore = testFullAlign(r,c,i,c-1, myseqA, myseqB);
	 
	  if (fullScore > bestFullScore) {
	    bestFullScore = fullScore;
	    bestR = r;
	    bestC = c;
	    bestOldR = i;
	    bestOldC = c - 1;
	    
	    if (fullScore > 0.6 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.7 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.8 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }

	  }
	}

	else if (fullScore > secondBestScore) {
	  secondBestR = r;
	  secondBestC = c;
	  secondBestOldR = i;
	  secondBestOldC = c - 1;
	  secondBestScore = fullScore;
	}
      }
    }
    
    for (i = 0; i < c - 1; i++) {
      
      if (myseqA->ss[i] == myseqB->ss[r-1]) {

	fullScore = computeAngleScore(r,c,r-1,i,myseqA, myseqB) + computeOriginScore(r,c,r-1,i,myseqA, myseqB);

	if (fullScore > currentD) {
	  
	  fullScore = testFullAlign(r,c,r-1,i, myseqA, myseqB);

	  if (fullScore > bestFullScore) {
	    bestFullScore = fullScore;
	    bestR = r;
	    bestC = c;
	    bestOldR = r - 1;
	    bestOldC = i;
	    if (fullScore > 0.6 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.7 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.8 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	  }
	}
	else if (fullScore > secondBestScore) {
	  secondBestR = r;
	  secondBestC = c;
	  secondBestOldR = r - 1;
	  secondBestOldC = i;
	  secondBestScore = fullScore;
	}
      }
    }
  }
}

float computePenalty_ss(int r, int c, int fromR, int fromC) {

  float retvalue;

  retvalue = 0.0;

  if (r - fromR > 1) {    
    retvalue = penalty_ss[0] + ((float)(r - fromR - 1))*(penalty_ss[1]);
  }
  else if (c - fromC > 1) {
    retvalue = penalty_ss[2] + ((float)(c - fromC - 1))*(penalty_ss[3]);
  }
    
  return retvalue;

}




void align_ss(int full) {
  int r, c;
  float fullScore;
  for (r = 0; r < rowB; r++) {
    for (c = 0; c < colA; c++) {
      if (full == 1) {
	computeData_ss(r, c, &seqA, &seqB);
	fixMax(r, c);
      }
      else {
	computeData_onlyAngles(r,c, &seqA, &seqB);
      }
    }
  }

  if (full != 1) {
    fullScore = testFullAlign(secondBestR,secondBestC,secondBestOldR,secondBestOldC, &seqA, &seqB);
    
    if (fullScore > bestFullScore) {
      bestFullScore = fullScore;
      bestR = secondBestR;
      bestC = secondBestC;
      bestOldR = secondBestOldR;
      bestOldC = secondBestOldC;

    }	
  }

}




void report_ss(int full) {
  
  if (full == 1) {
#if SHOWOUTPUT 
    printf("\nMax Score %f\n", maxScore); 
#endif
    writePair_ss(maxIndex[0], maxIndex[1]);
  }
  else {
    printf("\nMax Full Score %f\n", bestFullScore); 
    printf("Columns: %d %d\tRows: %d %d\n", bestOldC, bestC, bestOldR, bestR);
    printf("Total Pairs = %d\n", tempTotalPairs);

    globalCalledAddPair = 4;
    
    init_pairs();
    
    add_pair(seqA.origin[bestC], seqB.origin[bestR], 1);
    add_pair(seqA.end[bestC], seqB.end[bestR], 1);
    add_pair(seqA.origin[bestOldC], seqB.origin[bestOldR], 1);
    add_pair(seqA.end[bestOldC], seqB.end[bestOldR], 1);
    

  }    

  free(data);

}

void printScores_ss() {
  
  int i;

  printf("Scores\n\n");
  for (i = 0; i < colA*rowB; i++) {
    printf("%f ", data[i].score);
    if ((i+1)%colA == 0) 
      printf("\n");
  }
}


